#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2013  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from .common import DAEMON_THREADS, get_ip_address, unbound_address

from .Base import BaseServer

import errno
import logging
import socket
import socketserver
import struct
import time

MULTICAST_GROUP = '224.4.4.4'
MULTICAST_PORT = 32768
MULTICAST_INTERVAL = 300

################################## Functions ###################################
def multicast(message, group, port, bind_address=""):
  try:
    message = message.encode('UTF-8')
  except AttributeError:
    pass
  sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
  sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
  if bind_address:
    sock.bind((bind_address, 0))
  sock.sendto(message, (group, port))



################################### Threads ####################################

def multicast_announcer(
  message,
  group=MULTICAST_GROUP,
  port=MULTICAST_PORT,
  interval=MULTICAST_INTERVAL,
  delay=1,
  interfaces=None
):
  '''
  Periodically announce presence via multicast.
  '''
  time.sleep(delay)
  # No point in looping if all interfaces will be used.
  if not interfaces or any(unbound_address(i) for i in interfaces):
    interfaces = ['']
  while True:
    logging.info('announcing presence by multicast (group: {})'.format(group))
    for interface in interfaces:
      # Do this here to ensure that changes are detected if an interface address
      # changes.
      bind_address = None
      if not unbound_address(interface):
        try:
          bind_address = get_ip_address(interface)
        except OSError as e:
#           if e.errno != errno.ENODEV:
#             raise e
#           else:
          pass

      # If the "interface" is an IP address then get_ip_address will return None.
      if not bind_address:
        bind_address = interface

      try:
        multicast(message, group, port, bind_address)
        if bind_address:
          logging.info('announcement sent via {}'.format(bind_address))
        else:
          logging.info('announcement sent via all interfaces')
      except socket.gaierror as e:
#         if e.errno == socket.EAI_NONAME:
#           continue
#         else:
          if bind_address:
            logging.error('announcement failed via {}: {}'.format(bind_address, e.strerror))
          else:
            logging.error('announcement failed: {}'.format(e.strerror))
          logging.error('if {} is an interface, make sure that it is configured'.format(interface))
    time.sleep(interval)



class MulticastServer(BaseServer, socketserver.UDPServer, object):

  # ThreadingMixIn attribute.
  daemon_threads = DAEMON_THREADS

  def __init__(
    self,
    server_address,
    handler,
    multicast_group,
    *args,
    **kwargs
  ):
    self.multicast_group = multicast_group
    if not unbound_address(server_address[0]):
      ip = get_ip_address(server_address[0])
      if ip:
        server_address = (ip, server_address[1])
    socketserver.UDPServer.__init__(self, server_address, handler, *args, **kwargs)



  def server_bind(self):
    if self.allow_reuse_address:
      self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    self.socket.bind(self.server_address)
    mreq = struct.pack(
      "4sl",
      socket.inet_aton(self.multicast_group),
      socket.INADDR_ANY
    )
    self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
#     self.server_address = self.get_address_and_port()



class TestHandler(socketserver.BaseRequestHandler, object):
  def handle(self):
    print("{}:\n{}".format(
      self.client_address[0],
      self.request[0]
    ))



def add_Multicast_argparse_groups(
  parser,
  multicast_address='0.0.0.0',
  multicast_port=MULTICAST_PORT,
  multicast_group=MULTICAST_GROUP,
  multicast_interval=MULTICAST_INTERVAL
):
  multicast_options = parser.add_argument_group(
    title="Multicast Options",
    description="Options that affect the behavior of the multicast (sub)server system.",
  )

  multicast_options.add_argument(
    "--multicast", action='store_true',
    help='Use multicasting to announce presence and detect other servers.',
  )

  multicast_options.add_argument(
    "--multicast-address", metavar='<interface|address>', default=multicast_address,
    help='The address to which to bind the multicast server socket. Default: %(default)s.',
  )

  multicast_options.add_argument(
    '--multicast-port', metavar='<port>', type=int, default=multicast_port,
    help='The multicast port. Default: %(default)s.',
  )

  multicast_options.add_argument(
    '--multicast-group', metavar='<group>', default=multicast_group,
    help='The multicast group. Default: %(default)s.',
  )

  multicast_options.add_argument(
    '--multicast-interval', metavar='<seconds>', type=int, default=multicast_interval,
    help='The multicast announcement interval. Default: %(default)s.',
  )

  multicast_options.add_argument(
    "--multicast-interface", metavar='<interface|address>', dest="multicast_interfaces",
    default=[], action='append',
    help='The interface or address through which to announce presence with multicast packets. If not given, all interfaces are used.',
  )
  return parser



if __name__ == "__main__":
  server = MulticastServer(
    ('', MULTICAST_PORT),
    TestHandler,
    MULTICAST_GROUP
  )
  server.serve_forever()
