#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2011-2013  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

'''
# Overview
The main server is an HTTP(S) server that serves files and provides various
interfaces (HTML, JSON, plaintext, etc.?). The main server may optionally run a
UDP subserver to listen for multicasts. When a multicast is received with a
valid datagram, the sending server is added to the server pool and a direct
reply is sent via an HTTP POST to avoid multicast flooding.

The server also provides various functions via HTTP POST requests. These are
mostly used internally.
'''

from .common import (
  add_common_argparse_groups,
  configure_logging,
  DAEMON_THREADS,
  format_seconds,
  get_name,
  run,
  ServerError,
  serverpath_to_localpath,
  serverpath_from_uripath,
  unbound_address,
  VERSION,
)

from .HTTPS import (
  add_HTTPS_argparse_groups,
)

from .Multicast import (
  add_Multicast_argparse_groups,
  MULTICAST_INTERVAL,
  MulticastServer,
  multicast_announcer,
)

from .Quickserve import (
  add_Quickserve_argparse_groups,
  QuickserveServer,
  QuickserveRequestHandler,
)

from .PageGenerators import DEFAULT_PAGE_GENERATORS, DEFAULT_MIMETYPE

import argparse
import itertools
import json
import logging
import os
import socket
import socketserver
import threading
import urllib.error
import urllib.parse


NAME = get_name(__file__, __package__)
VERSION_STRING = '{}/{}'.format(NAME, VERSION)

PORT = 8000
MULTICAST_GROUP = '224.3.45.66'
MULTICAST_PORT = 15680

TIMEOUT = 3

# TODO
# Consider how to rework this for HTTPS. Package integrity should be ensured
# through Pacman signatures and checksums, but private repositories might
# contain private/internal packages that should not be visible in transit.
DEFAULT_PEER_SCHEME = 'http'
PEER_FORMAT = '{}://{}:{:d}/'
NETLOC_PLACEHOLDER = 'ungabunga'

################################## Functions ###################################

def post_json(url, obj):
  '''
  Post a JSON-encoded object to the given url under the given name and interpret
  the response as a JSON object.
  '''
  data = {
    'json' : json.dumps(obj),
  }
  try:
    typ = obj['type']
  except KeyError:
    typ = 'unknown'
    print(json.dumps(obj))
  logging.info('POSTing to {} [type: {}]'.format(url, typ))
  data = urllib.parse.urlencode(data).encode('UTF-8')
  try:
    with urllib.request.urlopen(url, data=data, timeout=TIMEOUT) as f:
      return json.loads(f.read().decode())
  except (urllib.error.URLError, socket.timeout, ValueError) as e:
    try:
      logging.error('POST to {} failed [reason: {}]'.format(url, e.reason))
    except AttributeError:
      logging.error('POST to {} failed [error: {}]'.format(url, e))
  return None



def remote_file_check(url, paths, hops=0):
  if url[-1] != '/':
    url += '/'
  obj = {
    'type': 'file check',
    'paths': paths,
    'hops' : hops,
  }
  found = post_json(url, obj)
  try:
    for name, location in found.items():
      found[name] = urllib.parse.urljoin(url, location)
  except AttributeError:
    pass
  return found



def remote_directory_listing(url, path, hops=0):
  if url[-1] != '/':
    url += '/'
  obj = {
    'type': 'directory listing',
    'path': path,
    'hops': hops,
  }
  listing = post_json(url, obj)
  try:
    for name, entry in listing.items():
      for e in entry:
        try:
          e['href'] = urllib.parse.urljoin(url, e['href'])
        except KeyError:
          continue
      yield name, entry
  except AttributeError:
    pass


def announce_presence(url, scheme, port):
  if url[-1] != '/':
    url += '/'
  obj = {
    'type': 'nudge',
    'uri' : PEER_FORMAT.format(scheme, NETLOC_PLACEHOLDER, port),
  }
  return post_json(url, obj)



def replace_uri_netloc(url, netloc):
  url_data = urllib.parse.urlsplit(url)

  if not url_data.scheme:
    scheme = DEFAULT_PEER_SCHEME
  else:
    scheme = url_data.scheme

  if not url_data.port:
    port = PORT
  else:
    port = url_data.port

  return PEER_FORMAT.format(scheme, netloc, port), port



######################### MulticastQuickserveSubserver #########################
class MulticastQuickserveSubserver(MulticastServer):

  def __init__(self, main_server, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.main_server = main_server



##################### MulticastQuickserveSubRequestHandler #####################

class MulticastQuickserveSubRequestHandler(socketserver.BaseRequestHandler, object):
  def handle(self):
    self.server.main_server.handle_multicast(
      self.client_address,
      self.request[0].decode()
    )



########################## MulticastQuickserveServer ###########################
class MulticastQuickserveServer(QuickserveServer):
  '''
  Multicast-enabled Quickserve server.
  '''

  # Unique string to ensure protocol compatibility.
  MULTICAST_PREFIX = VERSION_STRING + ' '
  MAX_HOPS = 1

  def __init__(
    self,
    *args,
    **kwargs
  ):
    super().__init__(*args, **kwargs)

    self.peers = set()
    self.peer_lock = threading.Lock()

    if self.options.multicast:
      self.multicast_server = MulticastQuickserveSubserver(
        self,
        (self.options.multicast_address, self.options.multicast_port),
        MulticastQuickserveSubRequestHandler,
        self.options.multicast_group
      )
      self.multicast_server_thread = threading.Thread(
        target=self.multicast_server.serve_forever
      )
      self.multicast_server_thread.daemon = True #DAEMON_THREADS
      self.multicast_server_thread.start()

      self.multicast_announcer_thread = threading.Thread(
        target=multicast_announcer,
        args=(
          ('{}' + PEER_FORMAT).format(
            self.MULTICAST_PREFIX,
            self.get_scheme(),
            NETLOC_PLACEHOLDER,
            self.options.port
          ),
        ),
        kwargs={
          'group' : self.options.multicast_group,
          'port' : self.options.multicast_port,
          'interval' : self.options.multicast_interval,
          'interfaces' : self.options.multicast_interfaces,
        }
      )
      self.multicast_announcer_thread.daemon = True #DAEMON_THREADS
      self.multicast_announcer_thread.start()



  def shutdown(self):
    try:
      self.multicast_server.shutdown()
    except AttributeError:
      pass
    super().shutdown()



  def get_server_data(self, paths):
    address, port = self.get_address_and_port()
    address_header, address_string = self.get_address_header_and_string()

    multicast_address = self.options.multicast_address
    if unbound_address(multicast_address):
      multicast_address = 'all interfaces'

    data = [
      ('PID', os.getpid()),
      (address_header, address_string),
      ('Port', port),
    ]

    if self.options.multicast:
      if self.options.multicast_interfaces:
        multicast_interfaces = '\n'.join(self.options.multicast_interfaces)
      else:
        multicast_interfaces = 'all'
      data += [
        ('Multicast Address', multicast_address),
        ('Multicast Port', self.options.multicast_port),
        ('Multicast Group', self.options.multicast_group),
        ('Multicast Interval', format_seconds(self.options.multicast_interval)),
        ('Multicast Interfaces', multicast_interfaces),
      ]
    else:
      data += [('Multicast', False)]

    if self.options.peers:
      peers = '\n'.join(self.options.peers)
    else:
      peers = None

    data += [
      ('Filelist', getattr(self.options, 'filelist', None)),
      ('Filterlist', getattr(self.options, 'filterlist', None)),
      ('MOTD', getattr(self.options, 'motd', None)),
      ('Upload Directory', getattr(self.options, 'upload_directory', None)),
      ('Paths', paths),
      ('Static Peers', peers),
    ]

    return data


  def hide_peer_path(self, path):
    '''
    Return True if a peer path should not be accessible on the server.
    '''
    return False



  def notify_peers(self, peers=None):
    if not peers:
      peers = self.options.peers
    for uri in peers:
      self.log_message('announcing presence via POST to {}'.format(uri))
      announce_presence(uri, self.get_scheme(), self.options.port)



  def multicast_filter(self, address):
    for ip in self.local_ips():
      if address == (ip, self.options.port):
        return True
    return False



  def maybe_add_peer(self, uri, origin):
    if uri in self.options.peers:
      self.log_message('static peer {}'.format(uri))
      return True
    with self.peer_lock:
      if uri not in self.peers:
        self.peers.add(uri)
        self.log_message('added {} ({})'.format(uri, origin))
      return True



  def handle_multicast(self, client_address, data):
    l = len(self.MULTICAST_PREFIX)
    if data[:l] == self.MULTICAST_PREFIX:
      # TODO
      # Catch other errors here
      try:
        url = data[l:]
        url, port = replace_uri_netloc(data[l:], client_address[0])
      except ValueError:
        self.log_error('invalid multicast message from {}'.format(client_address[0]))
        return
      address = (client_address[0], port)
      if not self.multicast_filter(address):
        if self.maybe_add_peer(url, 'multicast'):
          self.log_message('POSTing to {}'.format(url))
          if announce_presence(url, self.get_scheme(), self.options.port):
            self.log_message('{} acknowledged POST'.format(url))
          else:
            self.log_message('{} rejected POST'.format(url))
    else:
      self.log_error('unrecognized multicast message from {}'.format(client_address[0]))



  # None indicates an error.
  def respond_to_post(self, handler, obj):
    try:
      if obj['type'] == 'file check':
        found = dict()
        unfound = set()
        for serverpath in obj['paths']:
          path = self.resolve_path(serverpath)
          if path is None:
            unfound.add(serverpath)
          else:
            found[serverpath] = serverpath
        if obj['hops'] > 0:
          obj['hops'] -= 1
          # Make a copy to avoid locking while searching other server. A lock or
          # copy is necessary to avoid errors if a peer is removed by another thread
          # during the iteraction. The lock is released during searches to
          # avoid holding it for too long while waiting for responses.
          with self.peer_lock:
            peers = list(self.peers)
          for url in self.options.peers + peers:
            search_results = remote_file_check(url, sorted(unfound), hops=obj['hops'])
            if search_results is None:
              with self.peer_lock:
                try:
                  self.peers.remove(url)
                except KeyError:
                  return
            else:
              for path, location in search_results.items():
                found[path] = location
                try:
                  unfound.remove(path)
                except KeyError:
                  pass
              if not unfound:
                break
        return found

      elif obj['type'] == 'directory listing':
        serverpath = obj['path']
        local_listing = self.resolve_path(serverpath)
        if local_listing and isinstance(local_listing, dict):
          return local_listing
        else:
          return None

      elif obj['type'] == 'nudge':
        uri, port = replace_uri_netloc(obj['uri'], handler.client_address[0])
        self.maybe_add_peer(uri, 'POST')
        return True
    except (KeyError, ValueError, TypeError):
      pass
    return None



  def get_navlinks(self, handler, page_generator):
    navlinks = super().get_navlinks(handler, page_generator)
    navlinks.append((
      handler.unparse_path(
        page='peers',
      ),
      'peers',
    ))
    return navlinks



###################### MulticastQuickserveRequestHandler #######################
class MulticastQuickserveRequestHandler(QuickserveRequestHandler):
  '''
  Pacserve request handler.
  '''

  def version_string(self):
    return VERSION_STRING



  def handle_custom(self):
    page_gen = super().handle_custom()
    if not page_gen:
      return None

    try:
      if self.url_qs['page'] == ['peers']:
        peers = sorted(set(self.server.options.peers) | self.server.peers)
        peer_links = (page_gen.format_link(p) for p in peers)
        peer_list = page_gen.format_list(peer_links, ordered=True)
        peer_section = page_gen.format_section(
          'Peer List', content=peer_list, level=2
        )
        page_gen.send_page(self, peer_section, title='Peers')
    except KeyError:
      pass

    return page_gen




  def handle_unresolved(self):
    if not self.server.hide_peer_path(self.url_path):
      search_obj = {
        'type' : 'file check',
        'paths' : [self.url_path],
        'hops' : self.server.MAX_HOPS,
      }
      found = self.server.respond_to_post(self, search_obj)
      try:
        location = found[self.url_path]
        self.log_message('redirecting to {}'.format(location))
        self.redirect(location)
        return True
      except (KeyError, TypeError):
        pass
    return False



  # Keep in sync with parent.
  def do_authenticated_GET_or_HEAD(self):
    self.parse_path()

    page_gen = self.handle_custom()
    if not page_gen:
      return self.send_error(404)

    resolved = self.server.resolve_path(self.url_path)
    if resolved is None:
      if not self.handle_unresolved():
        self.send_error(404)
    elif isinstance(resolved, str):
      self.transfer_file(resolved)
    elif self.url_data.path[-1] != '/':
      url_data = self.url_data[:2] + (self.url_data.path + '/',) + self.url_data[3:]
      self.redirect(location=urllib.parse.urlunsplit(url_data))
    else:
      ###### Remote Entries/ #####
      if self.server.options.list_remote:
        with self.server.peer_lock:
          peers = list(self.server.peers)
        for url in self.server.options.peers + peers:
          for name, entries in remote_directory_listing(url, self.url_path):
            try:
              resolved[name].extend(entries)
            except KeyError:
              resolved[name] = list(entries)
      ###### /Remote Entries #####

      if self.use_index():
        index_path = self.server.resolve_path(
          urllib.parse.urljoin(
            self.url_data.path,
            self.server.options.index
          )
        )
        if index_path:
          self.transfer_file(index_path)
        else:
          page_gen.send_directory_listing(self, resolved)
      else:
        page_gen.send_directory_listing(self, resolved)



  def do_authenticated_POST(self):
    try:
      if self.headers['Content-Type'] != 'application/x-www-form-urlencoded':
        raise ValueError('not for me')
      content_length = int(self.headers['Content-Length'])
    except (KeyError, ValueError):
      return super().do_authenticated_POST()
    data = self.rfile.read(content_length)
    obj = None
    for field in data.split(b'&'):
      if field[:5] == b'json=':
        jobj = urllib.parse.unquote_plus(field[5:].decode())
        obj = json.loads(jobj)
        break
    if obj is None or 'type' not in obj:
      self.send_error(400)
      return
    response = self.server.respond_to_post(self, obj)
    self.transfer_json(response)



##################################### Main #####################################

def add_MulticastQuickserve_argparse_groups(parser):

  mcqs_options = parser.add_argument_group(
    title="MulticastQuickserve Options",
  )

  mcqs_options.add_argument(
    '--peer', dest='peers', metavar='<scheme>://<host>:<port>/', default=[], action='append',
    help='Static peers. Pass the option multiple times if necessary. Example: "http://10.0.0.2:8000/"'
  )

  mcqs_options.add_argument(
    '--list-remote', action='store_true',
    help='Include remote files in directory listings.'
  )

  return parser



def main(args=None):
  parser = argparse.ArgumentParser(
    description='%(prog)s - Quickserve with p2p support.',
  )
  parser.add_argument(
    'paths', metavar='<filepath>', nargs='*',
    help='The files and directories to share. These will appear with the same name in server root. Use the filelist option for more advanced features.',
  )
  parser = add_Quickserve_argparse_groups(parser)
  parser = add_MulticastQuickserve_argparse_groups(parser)
  parser = add_common_argparse_groups(parser, port=PORT)
  parser = add_HTTPS_argparse_groups(parser)
  parser = add_Multicast_argparse_groups(
    parser,
    multicast_port=MULTICAST_PORT,
    multicast_group=MULTICAST_GROUP,
    multicast_interval=MULTICAST_INTERVAL
  )
  args = parser.parse_args(args)

  address = (args.address, args.port)

  page_generators = dict()
  for mimetype, pagegen in DEFAULT_PAGE_GENERATORS.items():
    page_generators[mimetype] = pagegen()
  args.default_mimetype = DEFAULT_MIMETYPE

  handler = MulticastQuickserveRequestHandler
  server = MulticastQuickserveServer(
    address,
    handler,
    args,
    page_generators,
  )
  print(server)
  print("Press ctrl+C to exit.")
  server.notify_peers()
  server.serve_forever()



if __name__ == '__main__':
  configure_logging()
  run(main)
