#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2013  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import errno
import fcntl
import logging
import os
import posixpath
import signal
import socket
import struct
import sys
import threading
import time
import urllib.parse

VERSION = '2013.05.12'
AUTHOR = 'Xyne'
HOMEPAGE = 'http://xyne.archlinux.ca/projects/quickserve'

# True means that all threads are killed when the main thread exits.
DAEMON_THREADS = True

# Number of times to attempt interrupting with exit before sending SIGKILL.
SIGINTS_BEFORE_SIGKILL = 1
SIGINTS_RECEIVED = 0

DEFAULT_CHUNK_SIZE = 0x1000
DEFAULT_CHUNK_FACTOR = 0x10 #0x80
RFC_2822_TIME_FORMAT = '%a, %d %b %Y %H:%M:%S %Z'
ISO_8601_TIME_FORMAT = '%Y-%m-%d %H:%M:%S %Z'



################################ Generic Error #################################

class ServerError(Exception):
  pass


################################ Path Functions ################################
# These should make the various URI path manipulations OS-agnostic.


def serverpath_from_uripath(uripath):
  '''
  The path is assumed to have been extracted with urllib.parse.urlsplit
  '''
  path = posixpath.normpath(urllib.parse.unquote(uripath))
  while path[:3] == '../':
    path = path[3:]
  return path


# Compare http.server's "translate_path"
def serverpath_to_localpath(path, start=None):
  path = serverpath_from_uripath(path)
  words = [w for w in path.split('/') if w]
  if start is None:
    path = os.getcwd()
  else:
    path = start
  for word in words:
    drive, word = os.path.splitdrive(word)
    head, word = os.path.split(word)
    if word in (os.curdir, os.pardir):
      continue
    path = os.path.join(path, word)
  return path




def iterate_qs(qs):
  '''
  Iterate values from urllib.parse.parse_qs-parsed query strings.
  '''
  for q, vs in qs.items():
    for v in vs:
      yield q, v



def unparse_qs(qs):
  '''
  Recreate a query string, i.e. reverse urllib.parse.parse_qs.
  '''
  return '&'.join(
    '{}={}'.format(
      urllib.parse.quote_plus(q),
      urllib.parse.quote_plus(v)
    )for q, v in iterate_qs(qs)
  )


#################################### System ####################################
def get_ip_address(ifname):
  '''
  Return the IP address of an interface (if it has one).
  '''
  SIOCGIFADDR = 0x8915
  sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  try:
    return socket.inet_ntoa(
      fcntl.ioctl(
        sock.fileno(),
        SIOCGIFADDR,
        struct.pack(b'256s', ifname[:15].encode('UTF-8'))
      )[20:24]
    )
  except OSError as e:
    if e.errno in (errno.ENODEV, errno.EADDRNOTAVAIL):
      return None
    else:
      raise e



def get_name(f, p):
  '''Call with __file__ and __package__.'''
  return os.path.splitext(f[f.rindex(p):])[0].replace(os.sep, '.')


def unbound_address(ip):
  return ip in ('', '0.0.0.0')



#################################### Format ####################################
def format_size(size):
  """Format bytes for inferior humans."""
  if size < 0x400:
    return '{:d} B'.format(size)
  else:
    size = float(size) / 0x400
  for prefix in ('KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB'):
    if size < 0x400:
      return '{:0.02f} {}'.format(size, prefix)
    else:
      size /= 0x400
  return '{:0.02f} YiB'.format(size)



def format_time(t=None, UTC=True):
  '''Format UNIX time to a ISO 8601 format.'''
  if UTC:
    t = time.gmtime(t)
  else:
    t = time.localtime(t)
  return time.strftime(ISO_8601_TIME_FORMAT, t)



def format_seconds(s):
  '''Format seconds for inferior humans.'''
  string = ''
  for base, char in (
    (60, 's'),
    (60, 'm'),
    (24, 'h')
  ):
    s, r = divmod(s, base)
    if s == 0:
      return '{:d}{}{}'.format(r, char, string)
    elif r != 0:
      string += '{:02d}{}{}'.format(r, char, string)
  else:
    return '{:d}d{}'.format(s, string)




##################################### Main #####################################

def add_common_argparse_groups(parser, port=8000):
  listening_options = parser.add_argument_group(
    title="Server Address and Port",
    description="Configure the server's listening address and port."
  )
  listening_options.add_argument(
    "-a", "--address", metavar='<interface|address>', default='',
    help='Bind the server to this address. By default the server will listen on all interfaces.',
  )
  listening_options.add_argument(
    "-p", "--port", type=int, metavar='<port>', default=port,
    help='Set the server port (default: %(default)s)',
  )
  listening_options.add_argument(
    "--ipv6", action='store_true',
    help='Use IPv6.',
  )
  return parser



def kill_myself(signum, frame):
  os.kill(os.getpid(), signal.SIGKILL)



def handle_interrupt(signum, frame):
  '''
  Attempt to exit on SIGINT (ctrl+c). Bind the signal handler to kill_myself()
  for subsequent calls.
  '''
  global SIGINTS_BEFORE_SIGKILL
  global SIGINTS_RECEIVED

  if SIGINTS_RECEIVED >= SIGINTS_BEFORE_SIGKILL:
    sys.stderr.write('\nThe next SIGINT will send a SIGKILL.\n')
    signal.signal(signal.SIGINT, kill_myself)

  elif SIGINTS_RECEIVED >= 1:
    sys.stderr.write(
      '\n{:d} SIGINTS remaining before SIGKILL.\n'.format(SIGINTS_REMAINING)
    )
  SIGINTS_RECEIVED += 1
  sys.exit(os.EX_OK)


def run(main, args=None):
  signal.signal(signal.SIGINT, handle_interrupt)
  try:
    main()
  except KeyboardInterrupt:
    sys.exit(os.EX_OK)
  except OSError as e:
    if e.errno in (errno.EADDRINUSE, errno.ENODEV):
      sys.stderr.write('{}\n'.format(e))
      sys.exit(os.EX_OSERR)
    else:
      raise e
  except ServerError as e:
    sys.stderr.write('{}\n'.format(e))
    sys.exit(os.EX_SOFTWARE)



def configure_logging():
  logging.basicConfig(
    datefmt=ISO_8601_TIME_FORMAT,
    format='[%(asctime)s] %(levelname)s: %(message)s',
    level='INFO',
  )